#include <QtCore/qcoreapplication.h>
#include <QtCore/qdebug.h>
#include <QtCore/qmap.h>
#include <QtCore/qobject.h>
#include <QtCore/qtimer.h>
#include <QtCore/qstring.h>
#include <QtQml/qjsengine.h>
#include <QtQml/qjsvalue.h>

// Creates a proxy object that acts as if it were a normal object, but
// retrieves all properties from somewhere else.
static const QString jsCode = QStringLiteral(R"666(
     (function(handler){
        const target = {};
        const proxy = new Proxy(target, handler);

        // Calls ownKeys() on the proxy handler and getOwnPropertyDescriptor()
        // For every property. Since we don't see a TypeError, I assume that the
        // returned descriptor is valid. The result on the console looks correct.
        console.log("The props: " + Object.keys(proxy));

        // We can see that ownKeys() is called, but not getOwnPropertyDescriptor().
        // No further action is taken. I tried to trace the QML engine, but got lost.
        //
        for (var props in proxy) {
            // is not shown, why?
            console.log('Iteration');
        }

        // There is also the possibility to use for..of instead, since we don't have
        // to consider prototypes anyway, but then the proxy tries to get
        // Symbol.iterator from the proxy handler and I have no idea how to create
        // that in C++.
        // for (var props of proxy) {
        //     console.log('Iteration');
        // }
     })
)666");

class ProxyHandler: public QObject {
    Q_OBJECT
public:
    ProxyHandler()
    {
        m_props = {
          { "prop1", "value1" },
          { "prop2", "value2" }
        };
    }

    Q_INVOKABLE bool has(const QJSValue &target, const QJSValue &key)
    {
        Q_UNUSED(target);
        Q_ASSERT(key.isString());
        QString name = key.toString();
        bool result = m_props.contains(name);

        qDebug() << "ProxyHandler::has(" << name << "): " << result;
        return result;
    }

    Q_INVOKABLE QJSValue get(const QJSValue &target, const QJSValue &key, const QJSValue &receiver)
    {
        Q_UNUSED(target);
        Q_UNUSED(receiver);
        Q_ASSERT(key.isString());

        QString name = key.toString();
        QString value = m_props.value(name, QString());

        qDebug() << "ProxyHandler::get(" << name << "): " << value;
        return value;
    }

    Q_INVOKABLE QStringList ownKeys(const QJSValue &target)
    {
        Q_UNUSED(target);
        QStringList keys = m_props.keys();

        qDebug() << "ProxyHandler::ownKeys(): " << keys;
        return keys;
    }

    Q_INVOKABLE QJSValue getOwnPropertyDescriptor(QJSValue target, const QJSValue &key)
    {
        Q_ASSERT(key.isString());

        QString name = key.toString();
        QJSValue value = QJSValue(m_props.value(name, QString()));

        // I found the properties below by tracing QV4Proxy::getOwnPropertyDescriptor()
        QJSValue result = qjsEngine(this)->newObject();
        result.setProperty(QStringLiteral("configurable"), true);
        result.setProperty(QStringLiteral("enumerable"), true);
        result.setProperty(QStringLiteral("writable"), false);
        result.setProperty(QStringLiteral("value"), value);

        // The JS engine checks whether the real target has compatible attributes.
        // This looks surprising, but see https://stackoverflow.com/questions/59274975/javascript-proxy-target-object-conflicts-with-property-descriptors
        // The workaround is to create a new (empty) property on-the-fly and add it to the target object.
        target.setProperty(name, value);

        qDebug() << "ProxyHandler::getOwnPropertyDescriptor(" << name << "): " << value.toString();
        return result;
    }

    QMap<QString, QString> m_props;
};

void test() {
    QJSEngine engine;
    engine.installExtensions(QJSEngine::ConsoleExtension);
    QJSValue proxyTest = engine.evaluate(jsCode, QStringLiteral("proxy.js"));
    if (proxyTest.isError())
        qWarning() << "Error: " << proxyTest.toString();
    Q_ASSERT(!proxyTest.isError());

    QJSValue handler = engine.newQObject(new ProxyHandler());
    QJSValue result = proxyTest.call(QJSValueList{ handler });
    if (result.isError())
        qWarning() << "Error: " << result.toString();
    Q_ASSERT(!result.isError());

    QCoreApplication::exit(0);
}

int main(int argc, char *argv[])
{
    QCoreApplication app(argc, argv);
    QTimer::singleShot(0, &test);
    return app.exec();
}

#include "main.moc"
